"""
Tests brewblox_ctl.click_helpers
"""

import click

from brewblox_ctl import click_helpers

TESTED = click_helpers.__name__


@click.group(cls=click_helpers.OrderedGroup)
def cli_one():
    pass


@click.group(cls=click_helpers.OrderedGroup)
def cli_two():
    pass


@cli_one.command()
def z_cmd_one_one():
    """Starts with Z to check list order"""


@cli_one.command()
def cmd_one_two():
    pass


@cli_two.command()
def cmd_two_one():
    pass


@cli_two.command()
def _cmd_two_two():
    """Starts with _ to check list order"""


def test_ordered_commands():
    cli = click_helpers.OrderedCommandCollection(
        sources=[
            cli_one,
            cli_two,
        ])

    assert cli.list_commands(None) == [
        'z-cmd-one-one',
        'cmd-one-two',
        'cmd-two-one',
        '-cmd-two-two',
    ]



"""
Tests brewblox_ctl.utils
"""

from os import path
from subprocess import DEVNULL, PIPE, STDOUT, CalledProcessError

import pytest

from brewblox_ctl import testing, utils

TESTED = utils.__name__


@pytest.fixture
def mocked_ext(mocker):
    mocked = [
        'input',
        'getcwd',
        'getenv_',
        'path',
        'which',
        'machine',
        'run',
        'set_key',
    ]
    return {k: mocker.patch(TESTED + '.' + k) for k in mocked}


@pytest.fixture
def mocked_opts(mocker):
    opts = utils.ContextOpts()
    mocker.patch(TESTED + '.ctx_opts').return_value = opts
    return opts


def test_ctx_opts():
    # Will raise an error outside click context
    with pytest.raises(RuntimeError):
        utils.ctx_opts()


def test_confirm(mocked_ext):
    mocked_ext['input'].side_effect = [
        '',
        'flapjacks',
        'NNNNO',
        'YES',
    ]
    assert utils.confirm('what?')  # default empty
    assert utils.confirm('why?')  # yes
    assert mocked_ext['input'].call_count == 4


def test_select(mocked_ext):
    m = mocked_ext['input']
    m.return_value = 'answer'
    assert utils.select('question?') == 'answer'
    assert m.call_count == 1

    m.clear_mock()
    m.return_value = ''
    assert utils.select('other question?', 'three') == 'three'
    assert m.call_count == 2


def test_check_ok(mocked_ext):
    m = mocked_ext['run']
    assert utils.check_ok('whatever')
    m.side_effect = CalledProcessError(1, '')
    assert not utils.check_ok('really?')

    m.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        utils.check_ok('truly?')


def test_confirm_usb(mocked_ext):
    utils.confirm_usb()
    assert mocked_ext['input'].call_count == 1


def test_setenv(mocked_ext, mocked_opts):
    set_mock = mocked_ext['set_key']

    utils.setenv('key', 'val')
    set_mock.assert_called_with(path.abspath('.env'), 'key', 'val', quote_mode='never')

    utils.setenv('key', 'other', '.other-env')
    set_mock.assert_called_with('.other-env', 'key', 'other', quote_mode='never')

    mocked_opts.dry_run = True
    utils.setenv('key', 'val-dry')
    assert set_mock.call_count == 2


def test_path_exists(mocked_ext):
    mocked_ext['path'].exists.side_effect = [
        True,
        False,
    ]
    assert utils.path_exists('p1')
    assert not utils.path_exists('p2')


def test_command_exists(mocked_ext):
    m = mocked_ext['which']
    m.return_value = ''
    assert not utils.command_exists('wobberjacky')

    m.return_value = '/bin/promised-land'
    assert utils.command_exists('pizza')


@pytest.mark.parametrize('combo', [
    ('armv7hf', True),
    ('armv6hf', True),
    ('amd64', False),
    ('x86-64', False)
])
def test_is_pi(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.is_pi() is result


@pytest.mark.parametrize('combo', [
    ('armv7hf', False),
    ('armv6hf', True),
    ('amd64', False),
    ('x86-64', False)
])
def test_is_v6(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.is_v6() is result


def test_is_root(mocked_ext):
    assert utils.is_root()

    mocked_ext['run'].side_effect = CalledProcessError(1, 'permission denied')
    assert not utils.is_root()


def test_is_docker_user(mocked_ext):
    assert utils.is_docker_user()

    mocked_ext['run'].side_effect = CalledProcessError(1, 'not found')
    assert not utils.is_docker_user()


def test_is_brewblox_cwd(mocked_ext):
    mocked_ext['getenv_'].side_effect = [
        '',
        '1.2.3',
    ]
    assert not utils.is_brewblox_cwd()
    assert utils.is_brewblox_cwd()


def test_optsudo(mocker):
    m = mocker.patch(TESTED + '.is_docker_user')
    m.side_effect = [
        True,
        False,
    ]
    assert utils.optsudo() == ''
    assert utils.optsudo() == 'sudo '


@pytest.mark.parametrize('combo', [
    ('armv6', 'rpi-'),
    ('armv7hf', 'rpi-'),
    ('amd64', ''),
    ('deep-thought', ''),
])
def test_tag_prefix(combo, mocked_ext):
    machine, result = combo
    mocked_ext['machine'].return_value = machine
    assert utils.tag_prefix() == result


def test_docker_tag(mocker):
    mocker.patch(TESTED + '.getenv').return_value = 'value'
    mocker.patch(TESTED + '.tag_prefix').return_value = 'prefix-'

    assert utils.docker_tag() == 'prefix-value'
    assert utils.docker_tag('release') == 'prefix-release'


def test_docker_tag_err(mocker):
    mocker.patch(TESTED + '.getenv').return_value = None
    mocker.patch(TESTED + '.tag_prefix').return_value = 'prefix-'

    assert utils.docker_tag('release') == 'prefix-release'
    with pytest.raises(KeyError):
        utils.docker_tag()


def test_check_config(mocked_ext):
    mocked_ext['getenv_'].side_effect = [
        '1.2.3',
        '',
        '',
        '',
        '',
    ]
    mocked_ext['input'].side_effect = [
        '',
        'no',
    ]
    # version ok
    assert utils.check_config()
    # version nok, user ok
    assert not utils.check_config(required=False)
    # required version nok
    with pytest.raises(SystemExit):
        utils.check_config()
    # version nok, user nok
    with pytest.raises(SystemExit):
        utils.check_config(required=False)


def test_sh(mocker):
    m_run = mocker.patch(TESTED + '.run')
    m_secho = mocker.patch(TESTED + '.click.secho')
    m_opts = mocker.patch(TESTED + '.ctx_opts').return_value

    m_opts.dry_run = False
    m_opts.verbose = False

    # Single call
    utils.sh('do things')
    assert m_secho.call_count == 0
    assert m_run.call_count == 1
    m_run.assert_called_with('do things',
                             shell=True,
                             check=True,
                             universal_newlines=False,
                             stdout=None,
                             stderr=STDOUT)

    m_run.reset_mock()
    m_secho.reset_mock()

    # Unchecked call
    utils.sh('do naughty things', check=False)
    assert m_secho.call_count == 0
    assert m_run.call_count == 1
    m_run.assert_called_with('do naughty things',
                             shell=True,
                             check=False,
                             universal_newlines=False,
                             stdout=None,
                             stderr=DEVNULL)

    m_run.reset_mock()
    m_secho.reset_mock()

    # Captured call
    utils.sh('gimme gimme', capture=True)
    assert m_secho.call_count == 0
    assert m_run.call_count == 1
    m_run.assert_called_with('gimme gimme',
                             shell=True,
                             check=True,
                             universal_newlines=True,
                             stdout=PIPE,
                             stderr=STDOUT)

    m_run.reset_mock()
    m_secho.reset_mock()

    # Dry run
    utils.sh('invisible shenannigans', utils.ContextOpts(dry_run=True))
    assert m_secho.call_count == 1
    assert m_run.call_count == 0

    m_run.reset_mock()
    m_secho.reset_mock()

    # List of commands
    utils.sh(['do this', 'and this', 'and that'])
    assert m_secho.call_count == 0
    assert m_run.call_count == 3

    m_run.reset_mock()
    m_secho.reset_mock()

    # Generator
    def generate():
        yield 'first'
        yield 'second'

    m_run.reset_mock()
    utils.sh(generate(), utils.ContextOpts(verbose=True))
    assert m_secho.call_count == 2
    assert m_run.call_count == 2
    m_run.assert_any_call('first',
                          shell=True,
                          check=True,
                          universal_newlines=False,
                          stdout=None,
                          stderr=STDOUT)
    m_run.assert_called_with('second',
                             shell=True,
                             check=True,
                             universal_newlines=False,
                             stdout=None,
                             stderr=STDOUT)


def test_logs(mocker):
    m_opts = mocker.patch(TESTED + '.ctx_opts').return_value
    m_secho = mocker.patch(TESTED + '.click.secho')

    m_opts.quiet = True
    utils.info('test')
    assert m_secho.call_count == 0
    utils.warn('warning')
    assert m_secho.call_count == 1
    utils.error('error')
    assert m_secho.call_count == 2

    m_opts.quiet = False
    utils.info('test')
    assert m_secho.call_count == 3
    utils.warn('warning')
    assert m_secho.call_count == 4
    utils.error('error')
    assert m_secho.call_count == 5


def test_load_ctl_lib(mocker):
    m_sudo = mocker.patch(TESTED + '.optsudo')
    m_sh = mocker.patch(TESTED + '.sh')
    m_getenv = mocker.patch(TESTED + '.getenv')

    m_sudo.return_value = 'SUDO '
    m_sh.side_effect = testing.check_sudo
    m_getenv.return_value = 'release'

    utils.load_ctl_lib()
    assert m_sh.call_count == 7

    m_sh.reset_mock()
    m_sh.side_effect = None  # remove check_sudo
    m_sudo.return_value = ''
    utils.load_ctl_lib()
    assert m_sh.call_count == 6

    m_sh.reset_mock()
    m_getenv.return_value = None
    with pytest.raises(KeyError):
        utils.load_ctl_lib()
    assert m_sh.call_count == 0



"""
Tests brewblox_ctl.commands.docker
"""


from unittest.mock import call

import pytest

from brewblox_ctl.commands import docker
from brewblox_ctl.testing import check_sudo, invoke

TESTED = docker.__name__


@pytest.fixture
def m_utils(mocker):
    m = mocker.patch(TESTED + '.utils')
    m.optsudo.return_value = 'SUDO '
    return m


@pytest.fixture
def m_sh(mocker):
    m = mocker.patch(TESTED + '.sh')
    m.side_effect = check_sudo
    return m


def test_up(m_utils, m_sh):
    invoke(docker.up)
    m_sh.assert_called_once_with('SUDO docker-compose up -d --remove-orphans')


def test_down(m_utils, m_sh):
    invoke(docker.down)
    m_sh.assert_called_once_with('SUDO docker-compose down --remove-orphans')


def test_restart(m_utils, m_sh):
    invoke(docker.restart)
    m_sh.assert_has_calls([
        call('SUDO docker-compose down --remove-orphans'),
        call('SUDO docker-compose up -d'),
    ])


def test_follow(m_utils, m_sh):
    invoke(docker.follow, 'spark-one spark-two')
    m_sh.assert_called_with('SUDO docker-compose logs --follow spark-one spark-two')
    invoke(docker.follow)
    m_sh.assert_called_with('SUDO docker-compose logs --follow ')


def test_kill(m_utils, m_sh):
    invoke(docker.kill)
    m_sh.assert_called_once_with('SUDO docker rm --force $(SUDO docker ps -aq)', check=False)



"""
Tests brewblox_ctl.commands.env
"""


import pytest

from brewblox_ctl import const
from brewblox_ctl.commands import env
from brewblox_ctl.testing import invoke

TESTED = env.__name__


@pytest.fixture
def m_utils(mocker):
    m = mocker.patch(TESTED + '.utils')
    m.optsudo.return_value = 'SUDO '
    return m


def test_skip_confirm(m_utils):
    invoke(env.skip_confirm)
    m_utils.setenv.assert_called_with(const.SKIP_CONFIRM_KEY, 'true')

    invoke(env.skip_confirm, ['FALSE'])
    m_utils.setenv.assert_called_with(const.SKIP_CONFIRM_KEY, 'false')

    invoke(env.skip_confirm, ['NOPE'], _err=True)
    assert m_utils.setenv.call_count == 2


def test_list_env(m_utils, mocker):
    m = mocker.patch(TESTED + '.dotenv.dotenv_values')
    m.return_value = {'k1': 'v1', 'k2': 'v2'}
    invoke(env.list_env)

    m.return_value = {}
    invoke(env.list_env)

    assert m.call_count == 2


def test_get_env(m_utils, mocker):
    m_utils.getenv.return_value = 'v1'
    invoke(env.get_env, 'k1')
    m_utils.getenv.assert_called_once_with('k1', '')


def test_set_env(m_utils):
    invoke(env.set_env, _err=True)
    invoke(env.set_env, ['k1', 'k2'])
    m_utils.setenv.assert_called_once_with('k1', 'k2')



"""
Tests brewblox_ctl.commands.install
"""


import pytest.__main__

from brewblox_ctl.commands import install
from brewblox_ctl.testing import check_sudo, invoke, matching

TESTED = install.__name__


@pytest.fixture(autouse=True)
def m_sleep(mocker):
    m = mocker.patch(TESTED + '.sleep')
    return m


@pytest.fixture
def m_utils(mocker):
    m = mocker.patch(TESTED + '.utils')
    m.optsudo.return_value = 'SUDO '
    m.docker_tag.side_effect = lambda v: v
    return m


@pytest.fixture
def m_sh(mocker):
    m = mocker.patch(TESTED + '.sh')
    m.side_effect = check_sudo
    return m


def test_install_short(m_utils, m_sh):
    m_utils.path_exists.return_value = False
    m_utils.command_exists.side_effect = [
        False,  # apt
        True,  # docker
        True,  # docker-compose
    ]
    invoke(install.install, '--use-defaults')
    assert m_sh.call_count == 3


def test_install_full(m_utils, m_sh):
    m_utils.path_exists.return_value = False
    m_utils.is_docker_user.return_value = False
    m_utils.command_exists.side_effect = [
        True,  # apt
        False,  # docker
        False,  # docker-compose
    ]
    invoke(install.install)
    assert m_sh.call_count == 7


def test_install_decline(m_utils, m_sh):
    m_utils.path_exists.return_value = False
    m_utils.is_docker_user.return_value = False
    m_utils.confirm.return_value = False
    m_utils.command_exists.side_effect = [
        True,  # apt
        False,  # docker
        False,  # docker-compose
    ]
    invoke(install.install, '--dir ./brewblox --no-reboot')
    print(m_utils.confirm.call_args_list)
    assert m_utils.confirm.call_count == 5

    m_utils.confirm.reset_mock()
    m_utils.command_exists.side_effect = [
        True,  # apt
        False,  # docker
        False,  # docker-compose
    ]

    invoke(install.install)


def test_install_existing_declined(m_utils, m_sh):
    m_utils.path_exists.return_value = True
    m_utils.is_docker_user.return_value = True
    m_utils.confirm.side_effect = [
        True,  # Use defaults
        False,  # No to overwrite
    ]
    m_utils.command_exists.side_effect = [
        True,  # apt
        False,  # docker
        False,  # docker-compose
    ]
    invoke(install.install)
    m_sh.assert_not_called()


def test_install_existing_continue(m_utils, m_sh):
    m_utils.path_exists.return_value = True
    m_utils.confirm.side_effect = [
        True,  # ok using dir
        True,  # continue existing dir
    ]
    m_utils.is_docker_user.return_value = True
    m_utils.command_exists.side_effect = [
        False,  # apt
        True,  # docker
        True,  # docker-compose
    ]
    invoke(install.install, '--no-use-defaults')
    m_utils.confirm.assert_called_with(matching(r'.*brewblox already exists.*'))
    assert m_sh.call_count == 3
    m_sh.assert_called_with('sudo reboot')


def test_prepare_flasher(m_utils, m_sh):
    install.prepare_flasher('taggart', True)
    m_sh.assert_any_call('SUDO docker pull brewblox/firmware-flasher:taggart')
    m_sh.assert_any_call('SUDO docker-compose down')

    # No-op call
    m_sh.reset_mock()
    m_utils.path_exists.return_value = False

    install.prepare_flasher('taggart', False)
    m_sh.assert_not_called()


def test_run_flasher(m_utils, m_sh):
    install.run_flasher('taggart', 'do-stuff')
    m_sh.assert_called_with('SUDO docker run -it --rm --privileged brewblox/firmware-flasher:taggart do-stuff')


def test_flash(m_utils, m_sh):
    invoke(install.flash, '--release develop --pull')
    assert m_sh.call_count == 4
    m_sh.assert_called_with('SUDO docker run -it --rm --privileged brewblox/firmware-flasher:develop flash')


def test_bootloader(m_utils, m_sh):
    invoke(install.bootloader, '--release develop --pull --force')
    assert m_sh.call_count == 3
    m_sh.assert_called_with(
        'SUDO docker run -it --rm --privileged brewblox/firmware-flasher:develop flash-bootloader --force')


def test_wifi(m_utils, m_sh):
    invoke(install.wifi, '--release develop --pull')
    assert m_sh.call_count == 3



"""
Tests brewblox_ctl.__main__
"""

import pytest
from click.exceptions import UsageError

from brewblox_ctl import __main__ as main
from brewblox_ctl import utils

TESTED = main.__name__


@pytest.fixture(autouse=True)
def mocked_local_commands(mocker):
    m = mocker.patch(TESTED + '.local_commands')
    m.return_value = []
    return m


@pytest.fixture
def mocked_utils(mocker):
    return mocker.patch(TESTED + '.utils')


def test_check_lib(mocked_utils):
    mocked_utils.is_brewblox_cwd.side_effect = [
        True,  # run 1
        True,  # run 2
        True,  # run 3
        False,  # run 4
    ]
    mocked_utils.path_exists.side_effect = [
        False,  # run 1
        False,  # run 2
        True,  # run 3
    ]
    mocked_utils.confirm.side_effect = [
        True,  # run 1
        False,  # run 2
    ]

    main.check_lib()  # ok
    main.check_lib()  # user nok
    main.check_lib()  # files exist
    main.check_lib()  # not a brewblox cwd

    assert mocked_utils.load_ctl_lib.call_count == 1


def test_main(mocked_utils, mocker):
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = False
    mocked_utils.getenv.return_value = None
    mocked_utils.ContextOpts = utils.ContextOpts

    main.main(['--help'])
    main.main(['env', 'get', 'USER'])


def test_is_root(mocked_utils):
    mocked_utils.is_root.return_value = True
    mocked_utils.is_v6.return_value = False
    with pytest.raises(SystemExit):
        main.main()


def test_is_v6(mocked_utils, mocker):
    mock_cli = mocker.patch(TESTED + '.click_helpers.OrderedCommandCollection')
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = True

    mocked_utils.confirm.side_effect = [
        False,
        True,
    ]

    with pytest.raises(SystemExit):
        main.main()

    main.main()
    assert mock_cli.return_value.call_count == 1


def test_exception(mocked_utils, mocker):
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = False
    mocked_utils.getenv.return_value = None

    with pytest.raises(SystemExit):
        main.main(['pancakes'])


def test_usage_hint(mocker, mocked_utils):
    mocked_utils.is_brewblox_cwd.return_value = True

    main.usage_hint(UsageError(''))
    assert mocked_utils.is_brewblox_cwd.call_count == 0

    main.usage_hint(UsageError('No such command'))
    assert mocked_utils.is_brewblox_cwd.call_count == 1

    mocked_utils.is_brewblox_cwd.return_value = False
    mocked_utils.path_exists.return_value = False

    main.usage_hint(UsageError('No such command'))

    mocked_utils.path_exists.return_value = True

    main.usage_hint(UsageError('No such command'))

    mock_cli = mocker.patch(TESTED + '.click_helpers.OrderedCommandCollection')
    mock_cli.return_value.side_effect = main.UsageError('No such command')
    mocked_utils.is_root.return_value = False
    mocked_utils.is_v6.return_value = False

    # call to getenv(BREWBLOX_DEBUG)
    mocked_utils.getenv.side_effect = [
        'true',
        None,
    ]

    with pytest.raises(UsageError):
        main.main(['nope'])

    with pytest.raises(SystemExit):
        main.main(['nope'])



"""
Tests brewblox_ctl.http
"""

import json
from unittest.mock import call, mock_open

import pytest

from brewblox_ctl.commands import http
from brewblox_ctl.testing import invoke

TESTED = http.__name__


@pytest.fixture
def mock_wait(mocker):
    return mocker.patch(TESTED + '.wait')


@pytest.fixture
def mock_requests(mocker):
    m = mocker.patch(TESTED + '.requests')
    for meth in http.METHODS:
        getattr(m, meth).return_value.text = meth + '-response'
    return m


@pytest.fixture
def mock_retry_interval(mocker):
    return mocker.patch(TESTED + '.RETRY_INTERVAL_S', 0.001)


def test_wait(mock_requests, mock_retry_interval):
    invoke(http.http, ['wait', 'url'])
    assert mock_requests.get.call_count == 1


def test_wait_timeout(mocker, mock_requests, mock_retry_interval):
    mocker.patch(TESTED + '.RETRY_COUNT', 5)
    mock_requests.get.return_value.raise_for_status.side_effect = http.ConnectionError

    result = invoke(http.http, ['wait', 'url'], _err=True)
    assert isinstance(result.exception, TimeoutError)
    assert mock_requests.get.call_count == 5


def test_http_wait(mock_requests, mock_wait):
    invoke(http.http, ['wait', 'url'])
    assert mock_wait.call_count == 1
    assert mock_requests.call_count == 0


def test_http_methods(mock_requests):
    invoke(http.http, ['get', 'url'])

    assert mock_requests.get.call_args_list == [
        call('url', headers={}, params={}, verify=False)
    ]
    assert mock_requests.post.call_count == 0

    invoke(http.http, ['post', 'url', '--pretty'])
    assert mock_requests.post.call_args_list == [
        call('url', headers={}, params={}, verify=False)
    ]


def test_http_body(mock_requests):
    body = {'var1': 1, 'var2': 'val'}
    invoke(http.http, ['post', 'url', '-d', json.dumps(body)])
    invoke(http.http, ['post', 'url', '-d', json.dumps(body), '--json-body=false'])

    assert mock_requests.post.call_args_list == [
        call('url', headers={}, params={}, verify=False, json=body),
        call('url', headers={}, params={}, verify=False, data=json.dumps(body))
    ]


def test_http_body_conflict(mock_requests):
    invoke(http.http, ['post', 'url', '-d', 'text', '-f', 'myfile.json'], _err=True)
    assert mock_requests.post.call_count == 0


def test_http_file_body(mocker, mock_requests):
    body = {'var1': 1, 'var2': 'val'}
    open_mock = mocker.patch(TESTED + '.open', mock_open(read_data=json.dumps(body)))

    invoke(http.http, ['post', 'url', '-f', 'file.json'])
    invoke(http.http, ['post', 'url', '-f', 'file.json', '--json-body=false'])

    assert mock_requests.post.call_args_list == [
        call('url', headers={}, params={}, verify=False, json=body),
        call('url', headers={}, params={}, verify=False, data=json.dumps(body))
    ]
    assert open_mock.call_args_list == [
        call('file.json'),
        call('file.json'),
    ]


def test_http_error(mock_requests):
    mock_requests.post.return_value.raise_for_status.side_effect = http.ConnectionError
    body = {'var1': 1, 'var2': 'val'}
    invoke(http.http, ['post', 'url', '-d', json.dumps(body)], _err=True)


def test_allow_http_error(mock_requests):
    mock_requests.post.return_value.raise_for_status.side_effect = http.ConnectionError
    body = {'var1': 1, 'var2': 'val'}
    invoke(http.http, ['post', 'url', '-d', json.dumps(body), '--allow-fail'])


def test_output(mock_requests):
    result = invoke(http.http, ['get', 'url'])
    assert result.stdout == 'get-response\n'

    result = invoke(http.http, ['get', 'url', '--quiet'])
    assert result.stdout == ''



"""
Tests brewblox_ctl.testing
"""

from unittest.mock import Mock

import pytest

from brewblox_ctl import testing

TESTED = testing.__name__


def test_invoke(mocker):
    m_result = mocker.patch(TESTED + '.CliRunner').return_value.invoke.return_value
    m_result.exception = None
    assert testing.invoke('arg1', 'arg2', other='kwarg1') == m_result

    with pytest.raises(AssertionError):
        testing.invoke(_err=RuntimeError)

    m_result.exception = RuntimeError()
    assert testing.invoke(_err=RuntimeError) == m_result
    with pytest.raises(RuntimeError):
        testing.invoke()


def test_matching():
    obj = testing.matching(r'.art')
    assert obj == 'cart'
    assert obj == 'part'

    mock = Mock()
    mock('fart')
    mock.assert_called_with(obj)


def test_check_sudo():
    testing.check_sudo('something else')
    testing.check_sudo(v for v in 'xyz')
    testing.check_sudo('sudo docker run')
    testing.check_sudo('SUDO docker-compose up -d')


@pytest.mark.parametrize('cmd', [
    'docker run',
    '   docker run',
    '\n\ndocker-compose kill',
    'fluffy bunnies; docker-compose down',
    'pretty kitties&&docker',
    'pretty kitties&&     docker',
    'cuddly puppies||docker-compose',
    'cuddly puppies||\tdocker-compose',
    'cuddly puppies||\ndocker-compose',
])
def test_check_sudo_caught(cmd):
    with pytest.raises(AssertionError):
        testing.check_sudo(cmd)
